from .attributes import Attributes, Classes, Elements
from .core import GenericElement
from .elements.basic import Page
from .elements.format import HorizontalRule, LineBreak
from .elements.image import Image
from .elements.links import Link
from .elements.lists import ListItem, OrderedList, UnorderedList
from .elements.styles import Div, Span
from .elements.tables import Data, Header, Row, Table
from .elements.text import (
    Heading1,
    Heading2,
    Heading3,
    Heading4,
    Heading5,
    Heading6,
    Paragraph,
)

__all__ = [
    "Attributes",
    "Classes",
    "Elements",
    "GenericElement",
    "Page",
    "HorizontalRule",
    "LineBreak",
    "Image",
    "Link",
    "ListItem",
    "OrderedList",
    "UnorderedList",
    "Div",
    "Span",
    "Data",
    "Header",
    "Row",
    "Table",
    "Heading1",
    "Heading2",
    "Heading3",
    "Heading4",
    "Heading5",
    "Heading6",
    "Paragraph",
]



"""
Contains code related to HTML attributes.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterator, Self, Type

from .utilities.validate import (
    get_type_name_string,
    is_element,
    is_valid_type,
    raise_for_type,
    sanitize_class_name,
    split_preserving_quotes,
    types_to_tuple,
)

if TYPE_CHECKING:
    from .types import AttributeMap, AttributeValue, ClassesType, Element


class Classes:
    """Class for managing classes for HTML elements."""

    DEFAULT_REPLACEMENTS: dict[str, str] = {" ": "-"}

    def __init__(self, *names: str) -> None:
        """Initializes the Classes object."""

        # Initialize instance variables
        self._classes = dict()
        self.reset_replacements()

        # Set the classes
        self.set(*names)

    @classmethod
    def from_string(cls, string: str) -> Self:
        """Creates a Classes object from a string."""
        return cls(*string.split(" "))

    @property
    def classes(self) -> dict[str:str]:
        """Gets the stored classes as a dictionary.

        Keys are the original class names, values are the sanitized class names.
        """
        return self._classes

    @property
    def replacements(self) -> dict[str, str]:
        """Gets the replacements dictionary.

        The replacements dictionary controls the replacement of specified
        characters in provided class names when they are sanitized.

        Keys are the characters to be replaced, values are the replacements.
        """
        return self._replacements

    @replacements.setter
    def replacements(self, replacements: dict[str, str]) -> None:
        """Sets the replacements dictionary."""
        self._replacements = replacements
        self.set(*self._classes.keys())

    def reset_replacements(self) -> None:
        """Resets the replacements dictionary to its default value."""
        self._replacements = self.DEFAULT_REPLACEMENTS

    def add(self, *names: str) -> None:
        """Adds classes to the list of classes.

        Note that duplicate classes will be ignored.
        """

        # Determine the existing sanitized names
        old_sanitized_names = list(self._classes.values())
        # Determine the new sanitized names
        new_original_names = names
        new_sanitized_names = [self._sanitize_name(name) for name in names]
        new_names = zip(new_original_names, new_sanitized_names)
        # Check that duplicate classes aren't being added
        filtered_names = {
            new_original_name: new_sanitized_name
            for new_original_name, new_sanitized_name in new_names
            if new_sanitized_name not in old_sanitized_names
        }
        # Update the classes
        self._classes.update(filtered_names)

    def set(self, *names: str) -> None:
        """Sets the list of classes."""
        method_name = self.set.__name__
        if not all(isinstance(i, str) for i in names):
            raise TypeError(
                f"Arguments passed to {method_name} must be strings"
            )
        self._classes = {arg: self._sanitize_name(arg) for arg in names}

    def remove(self, name: str) -> tuple[str, str]:
        """Removes a class from the list of classes.

        Returns the removed class (if it exists) as a tuple, otherwise None.
        The tuple is in the form of (class name, sanitized class name).
        """

        # Try removing the class by its original name
        try:
            return name, self._classes.pop(name)
        except KeyError:
            pass
        # Try removing the class by its sanitized name
        for original_name, sanitized_name in self._classes.items():
            if sanitized_name == self._sanitize_name(name):
                return original_name, self._classes.pop(original_name)
        # If the class was not found, raise an exception
        raise KeyError(f"Class '{name}' not found")

    def clear(self) -> None:
        """Clears the list of classes."""
        self._classes.clear()

    def _sanitize_name(self, name: str) -> str:
        """Converts a class string into a valid class name."""
        return sanitize_class_name(name, replacements=self._replacements)

    def construct(self) -> str:
        """Generates the class string."""
        return " ".join(self._classes.values())

    def __eq__(self, other: Any) -> bool:
        """Determines whether two Classes objects are equal.

        Since keys are only kept for historical reasons, equality is determined
        by comparing the values (e.g., sanitized class names) of the classes.
        """
        if isinstance(other, self.__class__):
            return set(self._classes.values()) == set(other._classes.values())
        elif isinstance(other, dict):
            return set(self._classes.values()) == set(other.values())
        return False

    def __bool__(self) -> bool:
        """Determines whether the instance is empty."""
        return len(self._classes) > 0

    def __str__(self) -> str:
        """Gets the string version of the object."""
        return self.construct()

    def __repr__(self) -> str:
        """Gets the string representation of the object."""
        arg_string = ", ".join(repr(c) for c in self._classes.keys())
        return f"{self.__class__.__name__}({arg_string})"


class Attributes:
    """Class for managing attributes for HTML elements."""

    def __init__(self, attributes: AttributeMap | None = None) -> None:
        """Initializes the Attributes object."""

        # Initialize instance variables
        self._attributes: AttributeMap = {"class": Classes()}

        # Set the attributes
        if attributes is not None:
            self.set(attributes)

    @classmethod
    def from_string(cls, string: str) -> Self:
        """Creates an Attributes object from a string."""
        attributes = dict()
        pairs = split_preserving_quotes(string)
        for pair in pairs:
            key, _, value = pair.partition("=")
            value = value.strip("'")
            if not value:
                value = True
            attributes[key] = value
        return cls(attributes)

    @property
    def attributes(self) -> AttributeMap:
        """Gets the stored attributes.

        Keys are the attribute names, values are the attribute values.
        """
        return self._attributes

    @property
    def classes(self) -> Classes:
        """Gets the stored classes."""
        return self._attributes["class"]

    @classes.setter
    def classes(self, classes: ClassesType) -> None:
        """Sets the stored classes.

        Valid data types for the classes property are:
        - A string
        - An instance of the Classes class
        Other data types will raise a TypeError.

        Note that a provided value with data type string will be assumed to be
        a class string, and thus the Classes.from_string method will be used
        to create the Classes object.
        """
        if isinstance(classes, str):
            classes = Classes.from_string(classes)
        elif not isinstance(classes, Classes):
            # Handle any invalid data types during conversion
            classes = Classes(classes)
        self._attributes["class"] = classes

    def add(self, attributes: AttributeMap) -> None:
        """Adds attributes to the list of attributes.

        Note that duplicate attributes will be ignored, and that the 'class'
        attribute will never be added since it is a special case (it will
        always exist).
        """
        self._attributes.update(
            {
                key: value
                for key, value in attributes.items()
                if key not in self._attributes
            }
        )

    def set(self, attributes: AttributeMap) -> None:
        """Sets the list of attributes."""
        if "class" in attributes and isinstance(attributes["class"], str):
            attributes["class"] = Classes.from_string(attributes["class"])
        elif "class" in attributes and attributes["class"] is None:
            attributes["class"] = Classes()
        elif "class" not in attributes:
            attributes["class"] = Classes()
        self._attributes = dict(attributes)

    def remove(self, name: str) -> None:
        """Removes attributes from the list of attributes.

        Will raise a KeyError if the attribute does not exist.
        """
        if name not in self._attributes:
            raise KeyError(f"Attribute '{name}' not found")
        elif name == "class":
            self._attributes["class"].clear()
        else:
            self._attributes.pop(name)

    def clear(self) -> None:
        """Clears the attributes of the HTML object."""
        self._attributes.clear()
        self._attributes["class"] = Classes()

    def construct(self) -> str:
        """Generates the attribute string."""
        pairs = []
        for key, value in self._attributes.items():
            # None and True values are boolean attributes, False will be ignored
            if (value is None) or (isinstance(value, bool) and value):
                pairs.append(f"{key}")
            elif isinstance(value, Classes) and not value:
                continue
            elif value:
                pairs.append(f"{key}='{value}'")
        return " ".join(pairs)

    def __getitem__(self, key: str) -> AttributeValue:
        """Gets an attribute from the Attributes object."""
        return self._attributes[key]

    def __setitem__(self, key: str, value: AttributeValue) -> None:
        """Sets an attribute in the Attributes object."""
        self._attributes[key] = value

    def __eq__(self, other: Any) -> bool:
        """Determines whether two Attributes objects are equal."""
        if isinstance(other, self.__class__):
            return self._attributes == other._attributes
        elif isinstance(other, dict):
            return self._attributes == other
        return False

    def __bool__(self) -> bool:
        """Determines whether the instance is empty."""
        has_added_attributes = len(self._attributes) > 1
        has_added_classes = bool(self.classes)
        return has_added_attributes or has_added_classes

    def __str__(self) -> str:
        """Gets the string version of the object."""
        return self.construct()

    def __repr__(self) -> str:
        """Gets the string representation of the object."""
        if not bool(self):
            return f"{self.__class__.__name__}()"
        return f"{self.__class__.__name__}(attributes={self._attributes!r})"


class Elements:
    """Class for managing HTML elements."""

    def __init__(self, *elements: Element) -> None:
        """Initializes the Elements object."""

        # Validate data types of elements
        if not all(is_element(e) for e in elements):
            raise TypeError("Elements must be strings or builder objects")

        # Initialize instance variables
        self._elements: list[Element] = []
        self._max_elements: int | None = None
        self._valid_types: tuple[Type, ...] | None = None

        # Set the elements
        self.set(*elements)

    @property
    def elements(self) -> list[Element]:
        """Gets the stored elements."""
        return self._elements

    @property
    def max_elements(self) -> int | None:
        """Gets the maximum number of elements."""
        return self._max_elements

    @max_elements.setter
    def max_elements(self, value: int | None) -> None:
        """Sets the maximum number of elements."""
        is_integer = isinstance(value, int) and not isinstance(value, bool)
        is_none = value is None
        current_elements = len(self._elements)
        property_name = Elements.max_elements.fget.__name__
        if not is_integer and not is_none:
            raise TypeError(f"{property_name} must be an int or None")
        elif is_integer and value < 0:
            raise ValueError(f"{property_name} must be a positive integer")
        elif is_integer and value < current_elements:
            raise ValueError(
                f"{property_name} must be greater than or equal to the current "
                f"number of elements ({current_elements})"
            )
        self._max_elements = value

    @property
    def valid_types(self) -> tuple[Type, ...] | None:
        """Gets the valid types."""
        return self._valid_types

    @valid_types.setter
    def valid_types(
        self,
        valid_types: Type | list[Type] | tuple[Type, ...] | None,
    ) -> None:
        """Sets the valid types.

        A value of None will remove any type restrictions.
        """
        # Always allow removing type restrictions
        if valid_types is None:
            self._valid_types = None
            return
        # Check if current elements are not compatible with proposed types
        self._raise_if_current_elements_are_invalid(valid_types=valid_types)
        # Set the valid types
        if valid_types := types_to_tuple(valid_types):
            self._valid_types = valid_types
        else:
            self._valid_types = None

    def add(self, *elements: Element) -> None:
        """Adds elements to the list of elements."""
        if self.valid_types is not None:
            for element in elements:
                raise_for_type(element, expected_types=self.valid_types)
        self._raise_if_exceeds_max_elements(new_elements=len(elements))
        self._elements.extend(elements)

    def set(self, *elements: Element) -> None:
        """Sets the list of elements."""
        if self.valid_types is not None:
            for element in elements:
                raise_for_type(element, expected_types=self.valid_types)
        self._raise_if_exceeds_max_elements(
            new_elements=len(elements),
            ignore_current_elements=True,
        )
        self._elements = list(elements)

    def insert(self, index: int, element: Element) -> None:
        """Inserts the provided element at the specified index."""
        if self.valid_types is not None:
            raise_for_type(element, expected_types=self.valid_types)
        self._raise_if_exceeds_max_elements(new_elements=1)
        self._elements.insert(index, element)

    def update(self, index: int, element: Element) -> None:
        """Updates the provided element at the specified index."""
        if self.valid_types is not None:
            raise_for_type(element, expected_types=self.valid_types)
        self._elements[index] = element

    def remove(self, index: int) -> None:
        """Removes the element at the specified index."""
        del self._elements[index]

    def pop(self, index: int = -1) -> Element:
        """Pops and returns the element at the specified index."""
        return self._elements.pop(index)

    def clear(self) -> None:
        """Clears the list of elements."""
        self._elements.clear()

    def _raise_if_exceeds_max_elements(
        self,
        new_elements: int,
        ignore_current_elements: bool = False,
    ) -> None:
        """Raises an exception if the provided elements can not be added."""

        if self.max_elements is None:
            return
        elif ignore_current_elements:
            # Check if the number of new elements exceeds the max elements
            proposed_elements = new_elements
        else:
            # Check if the total number of elements exceeds the max elements
            proposed_elements = len(self._elements) + new_elements
        exceeds_max_elements = proposed_elements > self.max_elements
        if exceeds_max_elements:
            proposed_string = (
                "element" if proposed_elements == 1 else "elements"
            )
            raise ValueError(
                f"{proposed_elements} {proposed_string} would exceed the "
                f"maximum number of elements ({self.max_elements})"
            )

    def _raise_if_current_elements_are_invalid(
        self,
        valid_types: Type | list[Type] | tuple[Type, ...] | None,
    ) -> None:
        """Raises an exception if the current elements are not valid types."""
        valid_types = types_to_tuple(valid_types)
        for element in self._elements:
            if not is_valid_type(element, valid_types):
                type_names = get_type_name_string(valid_types)
                raise TypeError(
                    f"Types of current elements are not one of {type_names}"
                )

    def __getitem__(self, index: int) -> Element:
        """Gets the element at the specified index."""
        return self._elements[index]

    def __setitem__(self, index: int, element: Element) -> None:
        """Sets the element at the specified index."""
        self.update(index, element)

    def __delitem__(self, index: int) -> None:
        """Deletes the element at the specified index."""
        self.remove(index)

    def __iter__(self) -> Iterator[Element]:
        """Iterates over the elements in the list."""
        return iter(self._elements)

    def __eq__(self, other: Any) -> bool:
        """Determines whether two Elements objects are equal."""
        if isinstance(other, self.__class__):
            return self._elements == other._elements
        elif isinstance(other, list):
            return self._elements == other
        return False

    def __bool__(self) -> bool:
        """Determines whether the instance is empty."""
        return len(self._elements) > 0

    def __len__(self) -> int:
        """Gets the number of elements in the list."""
        return len(self._elements)

    def __str__(self) -> str:
        """Gets the string version of the object."""
        return "".join(str(e) for e in self._elements)

    def __repr__(self) -> str:
        """Gets the string representation of the object."""
        contents = ", ".join(repr(e) for e in self._elements)
        return f"{self.__class__.__name__}({contents})"



"""
Contains code for all link-related HTML elements.
"""

from ..core import GenericElement
from ..types import AttributesType, ClassesType, ElementsType


class Link(GenericElement):
    """Constructs an HTML link element."""

    def __init__(
        self,
        elements: ElementsType | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Link object."""

        # Initialize the builder
        super().__init__(
            tag="a",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )



"""
Contains types for the package.
"""

from typing import TypeAlias

from .attributes import Attributes, Classes, Elements
from .core import HTMLBuilder

# MARK: Builders

Builder: TypeAlias = HTMLBuilder

# MARK: Classes

ClassesType: TypeAlias = Classes | str

# MARK: Attributes

AttributeValue: TypeAlias = str | bool | None
AttributeMap: TypeAlias = dict[str, AttributeValue]
AttributesType: TypeAlias = Attributes | AttributeMap

# MARK: Elements

Element: TypeAlias = type[Builder] | str
ElementsType: TypeAlias = Elements | list[Element] | Element



"""
Contains code for all table-related HTML elements.
"""

from copy import deepcopy
from typing import Any, Iterator, Self

from ..attributes import Classes, Elements
from ..core import HTMLBuilder
from ..types import AttributesType, ClassesType
from ..utilities.optional import requires_modules
from ..utilities.validate import raise_for_type

# Import optional dependencies
try:
    import numpy as np
except ImportError:
    pass

try:
    import pandas as pd
except ImportError:
    pass


class Data(HTMLBuilder):
    """Constructs an HTML table data."""

    def __init__(
        self,
        data: Any | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
        is_header: bool = False,
    ) -> None:
        """Initializes the Data object."""

        # Initialize the builder
        super().__init__(
            elements=None,
            attributes=attributes,
            classes=classes,
        )
        self.tag = "td" if not is_header else "th"

        self.elements.max_elements = 1

        # Store instance variables
        self._is_header = is_header

        # Set the data
        if data is not None:
            self.set(data)

    @property
    def data(self) -> str:
        """Gets the data."""
        return self.elements[0] if self.elements else None

    def set(self, data: str) -> None:
        """Convenience wrapper for the self.elements.set method."""
        self.elements.set(data)

    def clear(self) -> None:
        """Convenience wrapper for the self.elements.clear method."""
        self.elements.clear()

    @property
    def is_header(self) -> bool:
        """Gets the is_header property."""
        return self._is_header

    @is_header.setter
    def is_header(self, value: bool) -> None:
        """Sets the is_header property."""
        self.tag = "td" if not value else "th"
        self._is_header = value

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        return super().construct()


class Row(HTMLBuilder):
    """Constructs an HTML table row."""

    def __init__(
        self,
        data: list[Data] | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Row object."""

        # Initialize the builder
        super().__init__(
            elements=Elements(*data) if data else Elements(),
            attributes=attributes,
            classes=classes,
        )
        self.tag = "tr"

        # Set the data
        if data is not None:
            self.set(*data)

    def add(self, *data: Data) -> None:
        """Convenience wrapper for the self.elements.add method."""
        self.elements.add(*data)

    def set(self, *data: Data) -> None:
        """Convenience wrapper for the self.elements.set method."""
        self.elements.set(*data)

    def insert(self, index: int, data: Data) -> None:
        """Convenience wrapper for the self.elements.insert method."""
        self.elements.insert(index, data)

    def update(self, index: int, data: Data) -> None:
        """Convenience wrapper for the self.elements.update method."""
        self.elements.update(index, data)

    def remove(self, index: int) -> None:
        """Convenience wrapper for the self.elements.remove method."""
        self.elements.remove(index)

    def pop(self, index: int = -1) -> Data:
        """Convenience wrapper for the self.elements.pop method."""
        return self.elements.pop(index)

    def clear(self) -> None:
        """Convenience wrapper for the self.elements.clear method."""
        self.elements.clear()

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        return super().construct()

    def __iter__(self) -> Iterator[Data]:
        """Iterates over the stored data."""
        return iter(self.elements)


class Header(Row):
    """Constructs an HTML table header row.

    Convenience wrapper around the Row class that will change any Data being
    added or modified to be header data.
    """

    def __init__(
        self,
        data: list[Data] | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Header object."""

        # Convert all data to header data
        if data is not None:
            for d in range(len(data)):
                try:
                    data[d].is_header = True
                except AttributeError:
                    raise TypeError(
                        f"Expected {Data.__name__} object, got "
                        f"{type(data[d]).__name__}"
                    )

        # Initialize the builder
        super().__init__(
            data=Elements(*data) if data else Elements(),
            attributes=attributes,
            classes=classes,
        )

    def add(self, *data: Data) -> None:
        """Convenience wrapper for the self.elements.add method."""
        for d in data:
            d.is_header = True
        self.elements.add(*data)

    def set(self, *data: Data) -> None:
        """Convenience wrapper for the self.elements.set method."""
        for d in data:
            d.is_header = True
        self.elements.set(*data)

    def insert(self, index: int, data: Data) -> None:
        """Convenience wrapper for the self.elements.insert method."""
        data.is_header = True
        self.elements.insert(index, data)

    def update(self, index: int, data: Data) -> None:
        """Convenience wrapper for the self.elements.update method."""
        data.is_header = True
        self.elements.update(index, data)


class Table(HTMLBuilder):
    """Constructs an HTML table."""

    def __init__(
        self,
        rows: list[Row] | None = None,
        header: Header | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Table object."""

        # Initialize the builder
        super().__init__(
            elements=Elements(*rows) if rows else Elements(),
            attributes=attributes,
            classes=classes,
        )
        self.tag = "table"

        # Set header and rows
        if rows is not None:
            self.set_rows(*rows)
        if header is not None:
            self.set_header(header)

    def _header_exists(self) -> bool:
        """Determines whether or not a header exists in the stored elements."""
        return bool(self.elements) and isinstance(self.elements[0], Header)

    @property
    def header(self) -> Header | Row | None:
        """Gets the header if one exists."""
        return self.elements[0] if self._header_exists() else None

    @header.setter
    def header(self, value: Header | Row | None) -> None:
        """Sets the header row."""
        self.set_header(value)

    @property
    def rows(self) -> list[Row]:
        """Gets the rows."""
        if self._header_exists():
            return self.elements[1:]
        return self.elements

    @rows.setter
    def rows(self, rows: list[Row]) -> None:
        """Sets the rows."""
        self.set_rows(*rows)

    def set_header(self, header: Header) -> None:
        """Sets the header row."""
        raise_for_type(header, expected_types=Header)
        if self._header_exists() and self.elements:
            self.elements.update(0, header)
        else:
            self.elements.insert(0, header)

    def clear_header(self) -> None:
        """Clears the header row."""
        if self._header_exists():
            self.elements.remove(0)

    def add_rows(self, *rows: Row) -> None:
        """Adds rows to the table."""
        for row in rows:
            raise_for_type(row, expected_types=Row)
            row.is_header = False
        self.elements.add(*rows)

    def set_rows(self, *rows: Row) -> None:
        """Sets rows for the table."""
        for row in rows:
            raise_for_type(row, expected_types=Row)
            row.is_header = False
        elements = [self.header] if self._header_exists() else []
        elements.extend(rows)
        self.elements.set(*elements)

    def clear_rows(self) -> None:
        """Clears the rows."""
        old_header = self.header
        self.elements.clear()
        if old_header is not None:
            self.set_header(old_header)

    def clear(self) -> None:
        """Clears both the header and the rows."""
        self.clear_header()
        self.clear_rows()

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        return super().construct()

    @classmethod
    @requires_modules("pandas", "numpy")
    def from_df(
        cls,
        df: pd.DataFrame,
        table_classes: ClassesType | None = None,
        header_classes: ClassesType | None = None,
        body_classes: ClassesType | None = None,
        attributes: AttributesType | None = None,
        alternating_rows: bool = True,
        columns_as_classes: bool = True,
    ) -> Self:
        """Creates an HTMLTable object from a pandas dataframe.

        Setting alternating_rows to True will add alternating classes to each
        row (excluding the header) for styling purposes. Even-numbered rows
        will have the class 'even' added and odd-numbered rows will have the
        class 'odd' added.
        """

        # Reset the index for the dataframe
        if alternating_rows:
            df.index = np.arange(1, len(df) + 1)

        # Create the body
        body: list[Row] = []
        for r, row in df.iterrows():
            data: list[Data] = []
            for i, item in enumerate(row):
                data_classes = Classes()
                if columns_as_classes:
                    data_classes.add(row.index[i])
                data.append(Data(item, classes=data_classes))
            html_row = Row(data, classes=deepcopy(body_classes))
            if alternating_rows:
                html_row.classes.add("odd" if r % 2 else "even")
            body.append(html_row)

        # Create the header
        header = Header(
            [Data(c) for c in df.columns],
            classes=deepcopy(header_classes),
        )

        # Create an instance with the results and return
        return cls(
            rows=body,
            header=header,
            attributes=attributes,
            classes=deepcopy(table_classes),
        )



"""
Contains code for all list-related HTML elements.
"""

from ..core import GenericElement
from ..types import AttributesType, ClassesType, Element, ElementsType


class ListItem(GenericElement):
    """Constructs an HTML list item element."""

    def __init__(
        self,
        elements: ElementsType | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the ListItem object."""

        # Initialize the builder
        super().__init__(
            tag="li",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class OrderedList(GenericElement):
    """Constructs an HTML ordered list."""

    def __init__(
        self,
        elements: list[ListItem] | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the OrderedList object."""

        # Initialize the builder
        super().__init__(
            tag="ol",
            elements=None,
            attributes=attributes,
            classes=classes,
        )

        self.elements.valid_types = ListItem

        # Set the elements
        if elements is not None:
            self.set(*elements)

    def add(self, *elements: Element) -> None:
        """Convenience wrapper for the self.elements.add method."""
        self.elements.add(*elements)

    def set(self, *elements: Element) -> None:
        """Convenience wrapper for the self.elements.set method."""
        self.elements.set(*elements)

    def insert(self, index: int, element: Element) -> None:
        """Convenience wrapper for the self.elements.insert method."""
        self.elements.insert(index, element)

    def update(self, index: int, element: Element) -> None:
        """Convenience wrapper for the self.elements.update method."""
        self.elements.update(index, element)


class UnorderedList(OrderedList):
    """Constructs an HTML unordered list."""

    def __init__(
        self,
        elements: list[ListItem] | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the UnorderedList object."""

        # Initialize the builder
        super().__init__(
            elements=elements,
            attributes=attributes,
            classes=classes,
        )
        self.tag = "ul"



"""
Contains code for all text-related HTML elements.
"""

from enum import Enum

from ..core import HTMLBuilder
from ..types import AttributesType, ClassesType


class TextType(Enum):
    P = "p"
    H1 = "h1"
    H2 = "h2"
    H3 = "h3"
    H4 = "h4"
    H5 = "h5"
    H6 = "h6"


class Text(HTMLBuilder):
    """Constructs HTML text."""

    def __init__(
        self,
        text: str | None = None,
        tag: TextType = TextType.P,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Text object."""

        # Initialize the builder
        super().__init__(
            elements=None,
            attributes=attributes,
            classes=classes,
        )
        self.tag = tag.value

        self.elements.max_elements = 1

        # Set the text
        if text is not None:
            self.set(text)

    @property
    def text(self) -> str:
        """Gets the text."""
        return self.elements[0] if self.elements else ""

    def set(self, text: str) -> None:
        """Convenience wrapper for the self.elements.set method."""
        self.elements.set(text)

    def clear(self) -> None:
        """Convenience wrapper for the self.elements.clear method."""
        self.elements.clear()

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        return super().construct()


class Paragraph(Text):
    """Constructs an HTML paragraph."""

    def __init__(
        self,
        text: str,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Paragraph object."""

        # Initialize the builder
        super().__init__(
            text,
            TextType.P,
            attributes=attributes,
            classes=classes,
        )


class Heading1(Text):
    """Constructs an HTML heading (size 1)."""

    def __init__(
        self,
        text: str,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Heading1 object."""

        # Initialize the builder
        super().__init__(
            text,
            TextType.H1,
            attributes=attributes,
            classes=classes,
        )


class Heading2(Text):
    """Constructs an HTML heading (size 2)."""

    def __init__(
        self,
        text: str,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Heading2 object."""

        # Initialize the builder
        super().__init__(
            text,
            TextType.H2,
            attributes=attributes,
            classes=classes,
        )


class Heading3(Text):
    """Constructs an HTML heading (size 3)."""

    def __init__(
        self,
        text: str,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Heading3 object."""

        # Initialize the builder
        super().__init__(
            text,
            TextType.H3,
            attributes=attributes,
            classes=classes,
        )


class Heading4(Text):
    """Constructs an HTML heading (size 4)."""

    def __init__(
        self,
        text: str,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Heading4 object."""

        # Initialize the builder
        super().__init__(
            text,
            TextType.H4,
            attributes=attributes,
            classes=classes,
        )


class Heading5(Text):
    """Constructs an HTML heading (size 5)."""

    def __init__(
        self,
        text: str,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Heading5 object."""

        # Initialize the builder
        super().__init__(
            text,
            TextType.H5,
            attributes=attributes,
            classes=classes,
        )


class Heading6(Text):
    """Constructs an HTML heading (size 6)."""

    def __init__(
        self,
        text: str,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Heading6 object."""

        # Initialize the builder
        super().__init__(
            text,
            TextType.H6,
            attributes=attributes,
            classes=classes,
        )



"""
Contains code for all style-related HTML elements.
"""

from ..core import GenericElement
from ..types import AttributesType, ClassesType, ElementsType


class Div(GenericElement):
    """Constructs an HTML div."""

    def __init__(
        self,
        elements: ElementsType | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Div object."""

        # Initialize the builder
        super().__init__(
            tag="div",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class Span(GenericElement):
    """Constructs an HTML span."""

    def __init__(
        self,
        elements: ElementsType | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Span object."""

        # Initialize the builder
        super().__init__(
            tag="span",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class Bold(GenericElement):
    """Constructs an HTML bold element."""

    def __init__(
        self,
        elements: ElementsType | str | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Bold object."""

        # Initialize the builder
        super().__init__(
            tag="b",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class Strong(GenericElement):
    """Constructs an HTML strong element."""

    def __init__(
        self,
        elements: ElementsType | str | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Strong object."""

        # Initialize the builder
        super().__init__(
            tag="strong",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class Italics(GenericElement):
    """Constructs an HTML italics element."""

    def __init__(
        self,
        elements: ElementsType | str | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Italics object."""

        # Initialize the builder
        super().__init__(
            tag="i",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class Emphasis(GenericElement):
    """Constructs an HTML emphasis element."""

    def __init__(
        self,
        elements: ElementsType | str | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Emphasis object."""

        # Initialize the builder
        super().__init__(
            tag="em",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class Underline(GenericElement):
    """Constructs an HTML underline element."""

    def __init__(
        self,
        elements: ElementsType | str | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Underline object."""

        # Initialize the builder
        super().__init__(
            tag="u",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class Strikethrough(GenericElement):
    """Constructs an HTML strikethrough element."""

    def __init__(
        self,
        elements: ElementsType | str | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Strikethrough object."""

        # Initialize the builder
        super().__init__(
            tag="s",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class Subscript(GenericElement):
    """Constructs an HTML subscript element."""

    def __init__(
        self,
        elements: ElementsType | str | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Subscript object."""

        # Initialize the builder
        super().__init__(
            tag="sub",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )


class Superscript(GenericElement):
    """Constructs an HTML superscript element."""

    def __init__(
        self,
        elements: ElementsType | str | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Superscript object."""

        # Initialize the builder
        super().__init__(
            tag="sup",
            elements=elements,
            attributes=attributes,
            classes=classes,
        )



"""
Contains core functionality for the package.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from .attributes import Attributes, Classes, Elements
from .utilities.optional import requires_modules
from .utilities.validate import is_element

if TYPE_CHECKING:
    from .types import AttributesType, ClassesType, Element, ElementsType

# Import optional dependencies
try:
    from bs4 import BeautifulSoup
    from bs4.formatter import HTMLFormatter
except ImportError:
    pass


class HTMLBuilder(ABC):
    """Base class for HTML Builder objects."""

    def __init__(
        self,
        elements: ElementsType | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the HTMLTable object."""

        # Standardize the elements
        if elements is not None:
            if is_element(elements):
                elements = Elements(elements)
            elif isinstance(elements, list):
                elements = Elements(*elements)
            elif not isinstance(elements, Elements):
                raise TypeError(
                    f"Invalid type {type(elements).__name__} for Elements"
                )

        # Initialize instance variables
        self._elements = elements if elements else Elements()
        self._attributes = attributes if attributes else Attributes()

        # Set the classes if any were provided
        if classes is not None:
            self._attributes.classes = classes

    @abstractmethod
    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        attributes_string = f" {self.attributes}" if self.attributes else ""
        html = f"<{self.tag}{attributes_string}>"
        for element in self.elements:
            html += f"{element}"
        html += f"</{self.tag}>"
        return html

    @requires_modules("bs4", "bs4.formatter")
    def prettify(self, indent: int = 2) -> str:
        """Generates HTML from the stored elements."""
        formatter = HTMLFormatter(void_element_close_prefix="", indent=indent)
        soup = BeautifulSoup(self.construct(), "html.parser")
        return soup.prettify(formatter=formatter)

    def save(self, filepath: str, prettify: bool = False) -> None:
        """Saves the HTML data to the specified filepath."""
        with open(filepath, "w", encoding="utf-8") as f:
            if not prettify:
                f.write(self.construct())
            else:
                try:
                    f.write(self.prettify())
                except ModuleNotFoundError:
                    f.write(self.construct())

    @property
    def elements(self) -> Elements:
        """Gets the stored elements."""
        return self._elements

    @property
    def attributes(self) -> Attributes:
        """Gets the stored attributes."""
        return self._attributes

    @property
    def classes(self) -> Classes | None:
        """Gets the stored classes."""
        return self._attributes.classes

    def __eq__(self, other: Any) -> bool:
        """Determines whether two HTMLBuilder objects are equal."""
        if isinstance(other, self.__class__):
            attributes_equal = self.attributes == other.attributes
            elements_equal = self.elements == other.elements
            return attributes_equal and elements_equal
        return False

    def __str__(self) -> str:
        """Gets a string version of the object."""
        return self.construct()

    def __repr__(self) -> str:
        """Gets a string representation of the object."""
        if not bool(self._attributes):
            return f"{self.__class__.__name__}()"
        return f"{self.__class__.__name__}(attributes={self._attributes!r})"


class GenericElement(HTMLBuilder):
    """Constructs a generic HTML element.

    This generic element class is useful for creating custom HTML elements that
    are simply containers for other sub-elements. Typically, the only difference
    between subclasses of GenericElement are the tags, but additional
    functionality may be implemented.

    It provides a convenient way to add, set, insert, update, remove, and pop
    elements, as well as a clear method to remove all elements.
    """

    def __init__(
        self,
        tag: str,
        elements: ElementsType | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the GenericElement object."""

        # Initialize the builder
        super().__init__(
            elements=elements,
            attributes=attributes,
            classes=classes,
        )
        self.tag = tag

    def add(self, *elements: Element) -> None:
        """Convenience wrapper for the self.elements.add method."""
        self.elements.add(*elements)

    def set(self, *elements: Element) -> None:
        """Convenience wrapper for the self.elements.set method."""
        self.elements.set(*elements)

    def insert(self, index: int, element: Element) -> None:
        """Convenience wrapper for the self.elements.insert method."""
        self.elements.insert(index, element)

    def update(self, index: int, element: Element) -> None:
        """Convenience wrapper for the self.elements.update method."""
        self.elements.update(index, element)

    def remove(self, index: int) -> None:
        """Convenience wrapper for the self.elements.remove method."""
        self.elements.remove(index)

    def pop(self, index: int = -1) -> Element:
        """Convenience wrapper for the self.elements.pop method."""
        return self.elements.pop(index)

    def clear(self) -> None:
        """Convenience wrapper for the self.elements.clear method."""
        self.elements.clear()

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        return super().construct()

    def __add__(self, other: Any) -> str:
        """Overloads the addition operator when the instance is on the left."""
        if isinstance(other, str):
            return self.construct() + other
        raise TypeError(
            f"Invalid type {type(other).__name__} for addition on "
            f"{self.__class__.__name__}; must be {str.__name__}"
        )

    def __radd__(self, other: Any) -> str:
        """Overloads the addition operator when the instance is on the right."""
        if isinstance(other, str):
            return other + self.construct()
        raise TypeError(
            f"Invalid type {type(other).__name__} for reverse addition on "
            f"{self.__class__.__name__}; must be {str.__name__}"
        )



"""
Contains code for optional dependencies.
"""

import importlib
from functools import wraps
from typing import Any, Callable


def module_exists(module_name: str) -> bool:
    """Determines whether or not a module exists."""
    try:
        importlib.import_module(module_name)
        return True
    except ImportError:
        return False


def requires_modules(*dependencies: str) -> Callable[[Callable], Callable]:
    """Raises an exception if any of the specified modules are not installed.

    Module names should be passed as separate string arguments.
    """

    def decorator(function: Callable) -> Callable:
        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Callable:
            # Check if all required dependencies are installed
            if missing := [d for d in dependencies if not module_exists(d)]:
                module_string = "module" if len(missing) == 1 else "modules"
                raise ModuleNotFoundError(
                    f"Function {function.__name__} requires the missing "
                    f"optional {module_string}: {', '.join(missing)}"
                )
            return function(*args, **kwargs)

        return wrapper

    return decorator



"""
Contains validation-related code for the package.
"""

import re
from typing import Any, Type

# MARK: Types and conversions


def is_builder(object: Any) -> bool:
    """Determines whether an object is a subclass of HTMLBuilder."""
    from ..types import Builder

    return issubclass(type(object), Builder) and type(object) is not Builder


def is_element(object: Any) -> bool:
    """Determines whether an object is a valid Element."""
    return is_builder(object) or isinstance(object, str)


def get_type_name_string(
    types: Type | list[Type] | tuple[Type, ...],
) -> str:
    """Formats a types tuple into a representation of its type names."""
    type_names = [t.__name__ for t in types_to_tuple(types)]
    return f"({', '.join(type_names)}{',' if len(type_names) == 1 else ''})"


def types_to_tuple(value: Any) -> tuple[Any, ...]:
    """Converts a value for expected types to a tuple if necessary."""
    if isinstance(value, tuple):
        return value
    elif isinstance(value, list):
        return tuple(value)
    elif not isinstance(value, Type):
        instance_type = type(value).__name__
        raise TypeError(f"Expected a type, got instance of {instance_type}")
    return (value,)


def is_valid_type(
    value: Any,
    expected_types: Type | list[Type] | tuple[Type, ...],
) -> bool:
    """Determines whether the input is of the expected type."""
    return isinstance(value, types_to_tuple(expected_types))


def raise_for_type(
    value: Any,
    expected_types: Type | list[Type] | tuple[Type, ...],
) -> None:
    """Determines whether the input is of the expected type."""
    expected_types = types_to_tuple(expected_types)
    if not is_valid_type(value, expected_types):
        raise TypeError(
            f"Got {type(value).__name__}, expected one of "
            f"{get_type_name_string(expected_types)}"
        )


# MARK: Classes


def is_valid_class_name(name: str) -> bool:
    """Determines whether a string is a valid HTML/CSS class name."""
    return re.match(r"^-?[_a-zA-Z]+[_a-zA-Z0-9-]*$", name) is not None


def sanitize_class_name(
    name: str,
    lower: bool = True,
    strip: bool = True,
    replacements: dict[str, str] | None = None,
) -> str:
    """Converts a class string into a valid class name."""
    if replacements is None:
        replacements = {" ": "-"}
    original_name = name
    name = name.lower() if lower else name
    name = name.strip() if strip else name
    for k, v in replacements.items():
        name = name.replace(k, v)
    if not is_valid_class_name(name):
        raise ValueError(
            f"Class name '{original_name}' (sanitized to '{name}') is invalid"
        )
    return name


# MARK: Attributes


def split_preserving_quotes(string: str) -> list[str]:
    """Splits an attribute string into a list of strings, preserving quotes."""
    return re.findall(r"[^'\s]+='[^']*'|\S+", string)



"""
Contains code for all top-level HTML classes.
"""

from ..core import HTMLBuilder
from ..types import Element, ElementsType


class Page(HTMLBuilder):
    """Constructs an HTML page."""

    def __init__(
        self,
        title: str,
        elements: ElementsType | None = None,
        lang: str | None = "en",
        charset: str | None = "UTF-8",
        stylesheets: list[str] | None = None,
    ) -> None:
        """Initializes the Page object.

        A non-empty title is required for valid HTML.

        The 'lang' attribute is "en" (english) by default, but can be changed
        to any valid language code. However, while it is highly recommended
        that this attribute is specified, this argument can be set to None and
        the attribute will not be added to the HTML.

        If no charset is provided, the default value of "UTF-8" will be used;
        however, if the charset is set to None, no charset meta tag will be
        added to the HTML.

        If no stylesheets are provided, no link tags will be added to the HTML.
        """

        # Initialize the builder
        super().__init__(
            elements=elements,
            attributes=None,
            classes=None,
        )
        self.tag: str = "html"

        # Initialize instance variables
        self.title: str = title
        self.lang: str | None = lang
        self.charset: str | None = charset
        self.stylesheets: list[str] = stylesheets

    @property
    def title(self) -> str:
        """Gets the title."""
        return self._title

    @title.setter
    def title(self, value: str) -> None:
        """Sets the title. Title must be a non-empty string."""
        if not (isinstance(value, str) and value):
            property_name = Page.title.fget.__name__
            raise TypeError(f"{property_name} must be a non-empty string")
        self._title = value

    @property
    def stylesheets(self) -> list[str]:
        """Gets the stylesheets."""
        return self._stylesheets

    @stylesheets.setter
    def stylesheets(self, value: list[str]) -> None:
        """Sets the stylesheets."""
        message = "stylesheets must be provided as a list of strings"
        if value is not None:
            if isinstance(value, list):
                if value and not all(isinstance(i, str) for i in value):
                    raise TypeError(message)
            else:
                raise TypeError(message)
        self._stylesheets: list[str] = value if value else []

    def add(self, *elements: Element) -> None:
        """Convenience wrapper for the self.elements.add method."""
        self.elements.add(*elements)

    def set(self, *elements: Element) -> None:
        """Convenience wrapper for the self.elements.set method."""
        self.elements.set(*elements)

    def insert(self, index: int, element: Element) -> None:
        """Convenience wrapper for the self.elements.insert method."""
        self.elements.insert(index, element)

    def update(self, index: int, element: Element) -> None:
        """Convenience wrapper for the self.elements.update method."""
        self.elements.update(index, element)

    def remove(self, index: int) -> None:
        """Convenience wrapper for the self.elements.remove method."""
        self.elements.remove(index)

    def pop(self, index: int = -1) -> Element:
        """Convenience wrapper for the self.elements.pop method."""
        return self.elements.pop(index)

    def clear(self) -> None:
        """Convenience wrapper for the self.elements.clear method."""
        self.elements.clear()

    def construct(self) -> str:
        """Generates HTML from the stored elements."""

        # Set up the page
        html = "<!DOCTYPE html>"

        # Open the tag
        attribute_string = f" lang='{self.lang}'" if self.lang else ""
        html += f"<{self.tag}{attribute_string}>"

        # Add the header
        html += "<head>"
        if self.charset:
            html += f"<meta charset='{self.charset}'>"
        html += f"<title>{self.title}</title>"
        for href in self.stylesheets:
            html += f"<link rel='stylesheet' href='{href}'>"
        html += "</head>"

        # Add the data
        html += "<body>"
        for element in self.elements:
            html += f"{element}"
        html += "</body>"

        # Close the tag and return the HTML
        html += f"</{self.tag}>"
        return html



"""
Contains code for all formatting-related HTML elements.
"""

from ..core import HTMLBuilder
from ..types import AttributesType, ClassesType


class Image(HTMLBuilder):
    """Constructs an HTML image."""

    def __init__(
        self,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Image object."""

        # Initialize the builder
        super().__init__(
            elements=None,
            attributes=attributes,
            classes=classes,
        )
        self.tag = "img"

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        attributes_string = f" {self.attributes}" if self.attributes else ""
        return f"<{self.tag}{attributes_string}>"
